import itertools
import tools

'''
This file contains various support for first order-predicate logic.

Note that all of the symbols, except of variable, should start with a lower-case letter. Only variable can start with upper-case letter.
Also note that symbols (names of variables, constants, functions, predicates) should not contain bracket.
'''


class Term:
    '''
    An interface for FOL term.
    '''

    def getVariables(self):
        '''
        Returns set of variables in this term.

        :rtype: set(Variable)
        '''
        raise NotImplementedError("Method not implemented.")

    def getConstants(self):
        '''
        Returns set of constants in this term.

        :rtype: set(Constant)
        '''
        raise NotImplementedError("Method not implemented.")

    def substitute(self, substitution):
        '''
        Returns the same term but substituted by the substitution represented by {Var: Term}.

        :type substitution: Dict({Var: Term})
        :rtype: Term
        '''
        raise NotImplementedError("Method not implemented.")


class Variable(Term):
    '''
    Represents FOL variable.
    '''

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def getVariables(self):
        result = set()
        result.add(self)
        return set(result)

    def substitute(self, substitution):
        if self in substitution:
            return substitution[self]
        return self

    def getnConstants(self):
        return set()

class Constant(Term):
    '''
    Represents FOL constant.
    '''

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return str(self.name)

    def getVariables(self):
        return set()

    def substitute(self, substitution):
        return self

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.name == other.name:
            return True
        return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return self.name.__hash__()

    def getConstants(self):
        constants = set()
        constants.add(self)
        return constants


class Function(Term):
    '''
    Reprsents FOL function.
    '''

    def __init__(self, name, arity):
        self.name = name
        self.arity = arity

    def __str__(self):
        return self.name + "/" + str(self.arity)


class FunctionTerm(Term):
    def __init__(self, function, terms):
        '''
        Creates and returns new function term which is constructed by a function applied to list of arguments.

        :type function: Function
        :type terms: List(Term)
        :rtype: FunctionTerm
        '''
        self.function = function
        self.terms = list(terms)

    def __str__(self):
        return self.function.name + "(" + ','.join(map(lambda term: str(term), self.terms)) + ")"

    def getVariables(self):
        return set(map(lambda term: term.getVariables(), self.terms))

    def substitute(self, substitution):
        return FunctionTerm(self.function, map(lambda term: term.substitute(substitution), self.terms))

    def getnConstants(self):
        return set(map(lambda term: term.getConstants(), self.terms))


class Predicate:
    def __init__(self, name, arity):
        '''
        Creates and returns new predicate by given name and arity. Length of the name should be longer than zero.

        :param name: string
        :param arity: int,>=0
        :rtype: Predicate
        '''
        self.name = name
        self.arity = arity

    def __str__(self):
        '''
        Get Prolog-like notation.

        :rtype: String
        '''
        return self.name + "/" + str(self.arity)


class Atom:
    def __init__(self, predicate, terms):
        '''
        Creates new atom given predicate and list of terms.

        :type predicate: Predicate
        :type terms: List(Terms)
        :rtype: Atom
        '''
        self.predicate = predicate
        self.terms = list(terms)
        self.arity = len(self.terms)

    def __str__(self):
        if 0 >= self.arity:
            return self.predicate.name
        return self.predicate.name + "(" + ','.join(map(lambda term: str(term), self.terms)) + ")"

    def getVariables(self):
        '''
        Returns set of variable sin this atom.

        :rtype: Set(Variable)
        '''
        return tools.unionSets(map(lambda term: set(term.getVariables()), self.terms))

    def substitute(self, substitution):
        '''
        Returns new atom obtained by applying the substitution on this atom.

        :type substitution: Dict({Var : Term})
        :rtype: Atom
        '''
        return Atom(self.predicate, map(lambda term: term.substitute(substitution), self.terms))

    def isGround(self):
        '''
        Returns true if the atom is ground.

        :rtype: bool
        '''
        return len(self.getVariables()) < 1

    def getConstants(self):
        '''
        Returns set of constants.

        :rtype: set(Constant)
        '''
        return tools.unionSets(map(lambda term: term.getConstants(),self.terms))

class Literal:
    def __init__(self, atom, positive=True):
        '''
        Creates and returns new literal from the atom. The literal is negation of the atom if positive is set to False

        :type atom: Atom
        :type positive: Bool
        :rtype: Literal
        '''
        self.atom = atom
        self.positive = positive

    def __str__(self):
        sign = ""
        if not self.positive:
            sign = "-"
        return sign + str(self.atom)

    def getPredicate(self):
        '''
        Return the predicate.

        :rtype: Predicate
        '''
        return self.atom.predicate

    def getVariables(self):
        '''
        Returns variables in this literal.

        :rtype: Set(Variable)
        '''
        return self.atom.getVariables()

    def substitute(self, substitution):
        '''
        Returns new literal obtained by applying the substitution on itself.

        :type substitution: Dict({Var : Term})
        :rtype: Literal
        '''
        return Literal(self.atom.substitute(substitution), self.positive)

    def isGround(self):
        '''
        Returns true if the atom is ground.

        :rtype: bool
        '''
        return self.atom.isGround()

    def getConstants(self):
        '''
        Returns set of constants.

        :rtype: set(Constant)
        '''
        return self.atom.getConstants()

class Clause:
    def __init__(self, literals):
        '''
        Creates a clause given the list of literals.

        :type literals: List(Literals)
        :rtype: Clause
        '''
        self.literals = list(literals)

    def __str__(self, endingDot=True):
        '''
        Set endingDot to False if you want to obtain string description of the clause without the ending dot.

        :type endingDot: bool
        :rtype: String
        '''
        dot = '.'
        if not endingDot:
            dot = ''
        if not self.literals:
            return '{}' + dot
        return ' | '.join(sorted(map(lambda literal: str(literal), self.literals))) + dot

    def getPredicates(self):
        '''
        Returns set of predicates in the clause.

        :rtype: Set(Predicate)
        '''
        return set(map(lambda literal: literal.getPredicate(), self.literals))

    def getVariables(self):
        '''
        Returns set of variables in the clause.

        :rtype: Set(Variable)
        '''
        if not self.literals:
            return set()
        return tools.unionSets(map(lambda literal: set(literal.atom.getVariables()), self.literals))

    def extend(self, literal):
        '''
        Creates and returns new clause by appending the given literal to the set of literals of this clause.

        :type literal: Literal
        :rtype: Clause
        '''
        terms = list(self.literals)
        terms.append(literal)
        return Clause(terms)

    def hasRedundantLiteral(self):
        '''
        Returns true if some literal occurs multiple times in the clause.

        :rtype: bool
        '''
        canonicSet = set(map(lambda literal: str(literal), self.literals))
        return not (len(canonicSet) == len(self.literals))

    def getCanonicString(self):
        '''
        Returns canonical string for the clause.

        :rtype: string
        '''
        variables = list(self.getVariables())
        count = len(variables)
        if count < 1:
            return str(self)
        minimal = None

        for permutation in itertools.permutations(list(map(lambda x: Constant("x"+str(x)), range(count))), count):
            substitution = {}
            for idx in range(count):
                substitution[variables[idx]] = permutation[idx]
            substituted = str(self.substitute(substitution))

            if minimal is None or substituted < minimal:
                minimal = substituted

        return minimal

    def substitute(self, substitution):
        '''
        Returns new clause by applying the substitution on itself.

        :type substitution: Dict({Var : Term})
        :rtype: Clause
        '''
        return Clause(map(lambda literal: literal.substitute(substitution), self.literals))

    def getNumberOfLiterals(self):
        return len(self.literals)

    def getNegativeLiterals(self):
        '''
        Returns set of negative literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: not literal.positive, self.literals))

    def getNegativeLiteralsAsClause(self):
        '''
        Returns subste of literals, only the negative ones, in a new clause.

        :rtype: Clause
        '''
        return Clause(self.getNegativeLiterals())

    def getPositiveLiterals(self):
        '''
        Returns set of positive literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: literal.positive, self.literals))

    def isGround(self):
        '''

        :rtype:
        '''
        return len(self.getVariables()) < 1

    def getGroundLiterals(self):
        '''
        Returns set of ground literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: len(literal.getVariables()) < 1, self.literals))

    def getNonGroundLiterals(self):
        '''
        Returns set of non-ground literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: len(literal.getVariables()) > 0, self.literals))

    def getNegativeNonGroundLiterals(self):
        '''
        Returns set of negative non-ground literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: not literal.positive, self.getNonGroundLiterals()))

    def getNonGroundAtoms(self):
        '''
        Returns set of non-ground atoms.

        :rtype: Set(Atom)
        '''
        return set(map(lambda literal: literal.atom, self.getNonGroundLiterals()))

    def getGroundAtoms(self):
        '''
        Returns set of ground atoms.

        :rtype: Set(Atom)
        '''
        return set(map(lambda literal: literal.atom, self.getGroundLiterals()))

    def isGround(self):
        '''
        Returns true if the clause is ground.

        :rtype: bool
        '''
        return all(map(lambda literal: literal.isGround(), self.literals))


class CNF:
    def __init__(self, clauses, variableFactory):
        '''
        Creates new CNF given list of clauses.

        :type clauses: List(Clauses)
        :type variableFactory: VariableFactory
        :rtype: CNF
        '''
        self.clauses = list(clauses)
        self.variableFactory = variableFactory

    def __str__(self):
        #return ' & '.join(map(lambda clause: '{' + clause.__str__(endingDot=False) + '}', self.clauses))
        return ' & '.join(map(lambda clause: '{' + clause.__str__(endingDot=False) + '}', self.clauses))

    def getPredicates(self):
        '''
        Returns set of predicates in the CNF.

        :rtype: Set(Predicate)
        '''
        return tools.unionSets(map(lambda clause: clause.getPredicates(), self.clauses))

    def copy(self):
        '''
        Returns copy of itself.

        :rtype: CNF
        '''
        return CNF(map(lambda clause: tools.replaceWithFreshVariables(clause, self.variableFactory), self.clauses),self.variableFactory)

    def getGroundLiterals(self):
        '''
        Returns set of ground literals.

        :rtype: Set(Literal)
        '''
        return tools.unionSets(map(lambda clause: clause.getGroundLiterals(), self.clauses))

    def getGroundAtoms(self):
        '''
        Returns set of ground atoms.

        :rtype: Set(Atom)
        '''
        return set(map(lambda literal: literal.atom, self.getGroundLiterals()))

    def getGroundNegativeLiteral(self):
        '''
        Returns set of ground negative literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: not literal.positive, self.getGroundLiterals()))

    def getGroundPositiveLiterals(self):
        '''
        Returns set of ground positive literals.

        :rtype: Set(Literal)
        '''
        return set(filter(lambda literal: literal.positive, self.getGroundLiterals()))

    def isGround(self):
        '''
        Returns True if the CNF is ground.

        :rtype: bool
        '''
        return all(map(lambda clause: clause.isGround(), self.clauses))

    def getNegativeLiterals(self):
        '''
        Returns set of negative literals.

        :rtype: Set(Literal)
        '''
        return tools.unionSets(map(lambda clause: clause.getNegativeLiterals(), self.clauses))

    def getPositiveLiterals(self):
        '''
        Returns set of positive literals.

        :rtype: Set(Literal)
        '''
        return tools.unionSets(map(lambda clause: set(clause.getPositiveLiterals()), self.clauses))

    def getLiterals(self):
        '''
        Returns set of literals.

        :rtype: Set(Literal)
        '''
        return tools.unionSets(map(lambda clause: set(clause.literals), self.clauses))


class Interpretation:
    """
    An interpretation is a set of ground atoms.

    Use .literals to get all ground literals in the interpretation.
    """

    def __init__(self, literals, variableFactory):
        '''
        Creates new interpretation given ground literals.

        :type literals: collection(Literal)
        :type variableFactory: VariableFactory
        :rtype: Interpretation
        '''
        self.literals = list(literals)
        self.variableFactory = variableFactory
        assert 0 == sum(map(lambda literal: len(literal.getVariables()), self.literals))
        assert 0 == len(set(filter(lambda literal: not literal.positive, self.literals)))

    def __str__(self):
        return '{' + ','.join(map(lambda literal: str(literal), self.literals)) + '}'

    def getPredicates(self):
        '''
        Returns set of predicates of the interpretation.

        :rtype: Set(Predicate)
        '''
        return set(map(lambda literal: literal.getPredicate(), self.literals))

    def isGround(self):
        '''
        Returns true if the interpretation is ground, which should be (just a check).

        :rtype: bool
        '''
        return all(map(lambda literal: literal.isGround(), self.literals))

    def getConstants(self):
        '''
        Returns set of constants.

        :rtype: set(Constant)
        '''
        return tools.unionSets(map(lambda literal: literal.getConstants(),self.literals))



